#include "fsck.h"

struct vd_context {
	uint64_t ino;
	struct nvfs_inode *inode;
	uint64_t n_blocks;
	const char *message;

	uint64_t *a;
	size_t al;
};

#define UH	0xFFFFFFFF00000000ULL

static int nde_compare(const void *nde1_t, const void *nde2_t)
{
	const struct nvfs_dir_entry *nde1 = nde1_t;
	const struct nvfs_dir_entry *nde2 = nde2_t;
	uint32_t namelen1 = nvfs_dir_entry_namelen(nde1);
	uint32_t namelen2 = nvfs_dir_entry_namelen(nde2);
	if (namelen1 != namelen2)
		return (int)namelen1 - (int)namelen2;
	return memcmp(nde1->name, nde2->name, namelen1);
}

static void free_empty(void *node)
{
	node = node + 0;
}

static bool validate_directory_entries(struct vd_context *ctx, struct nvfs_dir_head *h, uint32_t hash, unsigned hash_bits, void **tree)
{
	struct nvfs_dir_entry *nde = h->nde;
	struct nvfs_dir_entry *end = (void *)h + (1U << sb->log2_page_size);
	do {
		uint32_t namelen = nvfs_dir_entry_namelen(nde);
		size_t size = nvfs_dir_entry_needed_size(namelen);
		unsigned char type = nvfs_dir_entry_type(nde);
		if ((size_t)((void *)end - (void *)nde) < size) {
			ctx->message = "invalid leaf directory block";
			return false;
		}
		if (type) {
			void **ts;
			uint32_t h;
			if (memchr(nde->name, 0, namelen) ||
			    memchr(nde->name, '/', namelen)) {
				ctx->message = "invalid file name";
				return false;
			}
			if (namelen < 4) {
				size_t i;
				for (i = namelen; i < 4; i++) {
					if (nde->name[i] != 0) {
						ctx->message = "improperly padded name";
						return false;
					}
				}
			}
			h = nvfs_hash_name(nde->name, namelen);
			//fprintf(stderr, "hash: %08x (%.*s)\n", hash, namelen, nde->name);
			//fprintf(stderr, "hash: %08x (%08x, %u)\n", h, hash, hash_bits);
			if ((h & (UH >> hash_bits)) != hash) {
				ctx->message = "hash mismatch";
				return false;
			}
			ts = tsearch(nde, tree, nde_compare);
			if (*ts != nde) {
				ctx->message = "duplicate file name in directory";
				return false;
			}
		}
		nde = (void *)nde + size;
	} while (nde < end);
	return true;
}

static bool validate_directory_node(struct vd_context *ctx, struct nvfs_dir_head *h, uint32_t hash, unsigned hash_bits)
{
	uint64_t next;
	bool r = true;
	void *tree = NULL;
	unsigned long n = 0;

again:
	if (h->magic != NVFS_DIR_HEAD_MAGIC) {
		ctx->message = "invalid block in the chain";
		r = false;
		goto ret;
	}
	if (le64_to_cpu(h->self) != distance(h)) {
		//fprintf(stderr, "%lx %lx\n", le64_to_cpu(h->self), distance(h));
		ctx->message = "invalid self-pointer in the chain";
		r = false;
		goto ret;
	}
	if (le64_to_cpu(h->inode) != ctx->ino) {
		ctx->message = "invalid inode number in the chain";
		r = false;
		goto ret;
	}

	if (!validate_directory_entries(ctx, h, hash, hash_bits, &tree)) {
		r = false;
		goto ret;
	}

	next = le64_to_cpu(h->next);
	if (next) {
		if (n >= nvfs_max_chain_length(sb->log2_page_size)) {
			ctx->message = "the linear list is too long";
			r = false;
			goto ret;
		}
		n++;
		if (!verify_block(ctx->ino, &h->next, false)) {
			ctx->message = "invalid next block in the chain";
			r = false;
			goto ret;
		}
		h = ptr + next;
		append_array(&ctx->a, &ctx->al, next);
		ctx->n_blocks++;
		goto again;
	}

ret:
	tdestroy(tree, free_empty);
	return r;
}

static bool validate_directory_run(struct vd_context *ctx, __le64 *p, unsigned n, uint32_t hash, unsigned hash_bits)
{
	unsigned i = 0;
	while (i < n) {
		unsigned j, len;
		uint64_t v;
		unsigned bits, log2_len;
		uint32_t new_hash, new_hash_bits;

		for (j = i + 1; j < n; j++) {
			if (p[j] != p[i])
				break;
		}
		//fprintf(stderr, "%u - %u\n", i, j);

		len = j - i;

		if (len & (len - 1)) {
ret_false:
			ctx->message = "error in radix tree";
			return false;
		}

		if (i) {
			unsigned last_bit = i & -i;
			if (len > last_bit) {
				goto ret_false;
			}
		}

		if (!verify_block(ctx->ino, &p[i], false)) {
			ctx->message = "invalid block in tree";
			return false;
		}

		v = le64_to_cpu(p[i]);
		append_array(&ctx->a, &ctx->al, v);

		log2_len = __builtin_ctz(len);
		bits = __builtin_ctz(n >> log2_len);
		new_hash_bits = hash_bits + bits;
		new_hash = hash | (i >> log2_len << (32 - new_hash_bits));

		if (i >> log2_len << log2_len != i || new_hash_bits > 32) {
			ctx->message = "the directory is too deep";
			return false;
		}

		if (*(uint64_t *)(ptr + v) == NVFS_DIR_HEAD_MAGIC) {
			ctx->n_blocks++;
			if (!validate_directory_node(ctx, ptr + v, new_hash, new_hash_bits))
				return false;
		} else {
			if (len != 1) {
				ctx->message = "multiple pointers point to a non-leaf node";
				return false;
			}
			ctx->n_blocks++;
			if (!validate_directory_run(ctx, ptr + v, 1U << (sb->log2_page_size - 3), new_hash, new_hash_bits))
				return false;
		}

		i = j;
	}
	return true;
}

uint64_t alloc_dir_leaf(uint64_t ino)
{
	uint64_t b;
	struct nvfs_dir_head *ndh;
	b = alloc_block_or_inode(B_BLOCK);
	if (!b)
		return 0;
	b <<= sb->log2_page_size;
	ndh = ptr + b;
	ndh->magic = NVFS_DIR_HEAD_MAGIC;
	ndh->self = cpu_to_le64(b);
	ndh->next = cpu_to_le64(0);
	ndh->inode = cpu_to_le64(ino);
	ndh->nde->name_len_and_type = cpu_to_le32(nvfs_dir_entry_make_size((1UL << sb->log2_page_size) - NVFS_DIR_OFFSET) << NAME_LEN_OFFSET);
	return b;
}

bool validate_directory(uint64_t ino, struct nvfs_inode *inode)
{
	unsigned bank;
	struct vd_context ctx;
	bool rv = true;
	bool allocated_new = false;

	ctx.ino = ino;
	ctx.inode = inode;
	ctx.n_blocks = 0;

	init_array(&ctx.a, &ctx.al);

	bank = (le32_to_cpu(inode->i_flag) & NVFS_FLAG_SECOND_DIR_BANK) / NVFS_FLAG_SECOND_DIR_BANK;

	//ctx.message = "test";
	if (!validate_directory_run(&ctx, inode->i_radix[bank], NVFS_RADIX_BLOCKS, 0, 0) /*|| getenv("DIRTST")*/) {
		static struct qkey q;
		if (fsck_query(&q, Q_NORMAL, "directory inode %"PRIu64" has %s", "clear?", ctx.ino, ctx.message)) {
			uint64_t b;
			size_t i;
			for (i = 0; i < ctx.al; i++)
				bmp_op(ctx.a[i] >> sb->log2_page_size, B_CLEAR | B_BLOCK | B_MEM);
			b = alloc_dir_leaf(ino);
			if (!b) {
				static struct qkey q;
				if (fsck_query(&q, Q_NORMAL, "the disk is full - could not allocate directory leaf for inode %"PRIu64"", "clear?", ctx.ino)) {
					rv = false;
				}
			} else {
				for (i = 0; i < NVFS_RADIX_BLOCKS; i++)
					inode->i_radix[(le32_to_cpu(inode->i_flag) / NVFS_FLAG_SECOND_DIR_BANK) & 1][i] = cpu_to_le64(b);
			}
			inode->i_size = cpu_to_le64(1U << sb->log2_page_size);
			inode->i_blocks = cpu_to_le64(1U << (sb->log2_page_size - 9));
			allocated_new = true;
		}
	}

	fsck_free(ctx.a);

	if (!allocated_new) {
		if (ctx.n_blocks << sb->log2_page_size != le64_to_cpu(inode->i_size)) {
			static struct qkey q;
			if (fsck_query(&q, Q_PREEN, "directory inode %"PRIu64" has bad size (%"PRIu64" != %"PRIu64")", "fix?", ctx.ino, ctx.n_blocks << sb->log2_page_size, le64_to_cpu(inode->i_size))) {
				inode->i_size = cpu_to_le64(ctx.n_blocks << sb->log2_page_size);
			}
		}

		if (ctx.n_blocks << (sb->log2_page_size - 9) != le64_to_cpu(inode->i_blocks)) {
			static struct qkey q;
			if (fsck_query(&q, Q_PREEN, "directory inode %"PRIu64" has bad block count (%"PRIu64" != %"PRIu64")", "fix?", ctx.ino, ctx.n_blocks << (sb->log2_page_size - 9), le64_to_cpu(inode->i_blocks))) {
				inode->i_blocks = cpu_to_le64(ctx.n_blocks << (sb->log2_page_size - 9));
			}
		}
	}

	return rv;
}
